#!/usr/bin/env python3
# This file is part of GUFI, which is part of MarFS, which is released
# under the BSD license.
#
#
# Copyright (c) 2017, Los Alamos National Security (LANS), LLC
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation and/or
# other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#
# From Los Alamos National Security, LLC:
# LA-CC-15-039
#
# Copyright (c) 2017, Los Alamos National Security, LLC All rights reserved.
# Copyright 2017. Los Alamos National Security, LLC. This software was produced
# under U.S. Government contract DE-AC52-06NA25396 for Los Alamos National
# Laboratory (LANL), which is operated by Los Alamos National Security, LLC for
# the U.S. Department of Energy. The U.S. Government has rights to use,
# reproduce, and distribute this software.  NEITHER THE GOVERNMENT NOR LOS
# ALAMOS NATIONAL SECURITY, LLC MAKES ANY WARRANTY, EXPRESS OR IMPLIED, OR
# ASSUMES ANY LIABILITY FOR THE USE OF THIS SOFTWARE.  If software is
# modified to produce derivative works, such modified software should be
# clearly marked, so as not to confuse it with the version available from
# LANL.
#
# THIS SOFTWARE IS PROVIDED BY LOS ALAMOS NATIONAL SECURITY, LLC AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL LOS ALAMOS NATIONAL SECURITY, LLC OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
# OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
# OF SUCH DAMAGE.



import argparse
import math
import os
import subprocess
import sys
import time

import common
import bee_config

# hiding traceback
sys.tracebacklimit=0

# location of this file
PATH = os.path.realpath(__file__)

SECONDS_PER_DAY = str(24 * 60 * 60)

'''file size units specified in GNU find.'''
filesize = {'b' : 512,
            'c' : 1,
            'w' : 2,
            'k' : 1024,
            'M' : 1048576,
            'G' : 1073741824}

def validate_args(args):
    '''Validate known arguments here'''

    # beegfs-hive-index directories do not have source tree size/timestamps
    if args.type:
        if 'd' in args.type:
            if args.size or args.atime or args.mtime or args.ctime:
                raise Exception('size, atime, mtime, and ctime cannot be used with -type d')

# Used to parse all numeric arguments
# From man 1 find:
#
# Numeric arguments can be specified as
#
# +n     for greater than n,
#
# -n     for less than n,
#
# n      for exactly n.
#
def numeric_arg(n):
    try:
        float(n)
    except:
        raise argparse.ArgumentTypeError("%s is not a valid numeric argument" % n)

    op = None
    if n[0] == '+':
        op = '>'
        n = n[1:]
    elif n[0] == '-':
        op = '<'
        n = n[1:]
    else:
        op = '=='

    return op, n

def parse_size(size):
    unit = 'b'
    if size[-1] in filesize:
        unit = size[-1]
        size = size[:-1]

    comp, actual_size = numeric_arg(size)
    return comp, int(math.ceil(float(actual_size))) * filesize[unit]

def need_aggregation(args):
    return args.num_results or args.smallest or args.largest

def build_where(args, root_uid = 0, root_gid = 0):
    '''Build the WHERE clause'''
    where = []

    now = time.time()

    # if not ((os.geteuid() == root_uid) or (os.getegid() == root_gid)):
    #     where  += ['(uid == {})'.format(os.getuid())]
    #     # where  += ['(uid == {}) OR (gid in {})'.format(os.getuid(), '(' + str(os.getgroups())[1:-1] + ')')]

    if args.amin is not None:
        where += ['({} - atime) / 60 {} {}'.format(now, amin[0], amin[1]) for amin in args.amin]

    # if args.anewer is not None:

    if args.atime is not None:
        where += ['({} - atime) / {} {} {}'.format(now, SECONDS_PER_DAY, atime[0], atime[1]) for atime in args.atime]

    if args.cmin is not None:
        where += ['({} - ctime) / 60 {} {}'.format(now, cmin[0], cmin[1]) for cmin in args.cmin]

    # if args.cnewer is not None:

    if args.ctime is not None:
        where += ['({} - ctime) / {} {} {}'.format(now, SECONDS_PER_DAY, ctime[0], ctime[1]) for ctime in args.ctime]

    if args.empty == True:
        where += ['size == 0', '(type == \'f\') OR (type == \'d\')']

    if args.executable == True:
        where += ['(mode & 64) == 64'] # 0100 have to use decimal or hexadecimal value because sqlite3 doesn't support octal

    if args.false == True:
        where += ['0']

    if args.gid is not None:
        where += ['gid {} {}'.format(gid[0], gid[1]) for gid in args.gid]

    if args.group is not None:
        where += ['gid == {}'.format(args.group)]

    # if args.ilname is not None:

    # if args.iname is not None:
    #     where += ['name REGEXP \'(?i){}\''.format(args.iname)]

    if args.inum is not None:
        where += ['inode {} {}'.format(inum[0], inum[1]) for inum in args.inum]

    # Behaves in the same way as -iwholename. This option is deprecated, so please do not use it.
    # if args.ipath is not None:

    # if args.iregex is not None:

    # if args.iwholename is not None:

    if args.links is not None:
        where += ['nlink {} {}'.format(links[0], links[1]) for links in args.links]

    # if args.lname is not None:
    #     where += ['type == \'f\'', 'linkname GLOB \'(?i){}\''.format(args.lname)]

    if args.mmin is not None:
        where += ['({} - mtime) / 60 {} {}'.format(now, mmin[0], mmin[1]) for mmin in args.mmin]

    if args.mtime is not None:
        where += ['({} - mtime) / {} {} {}'.format(now, SECONDS_PER_DAY, mtime[0], mtime[1]) for mtime in args.mtime]

    if args.name is not None:
        where += [' OR '.join(['(name GLOB \'{}\')'.format(name) for name in args.name])]

    if args.newer is not None:
        where += ['mtime > {}'.format(args.newer.st_mtime)]

    # if args.newerXY is not None:

    # if args.nouser == True:

    # # not handled here
    # if args.path is not None:

    # if args.perm is not None:

    if args.readable == True:
        where += ['(mode & 256) == 256'] # 0400 have to use decimal or hexadecimal value because sqlite3 doesn't support octal

    # # not handled here
    # if args.regex is not None:

    if args.samefile is not None:
        where += ['inode == {}'.format(args.samefile.st_ino)]

    if args.size is not None:
        where += ['size {} {}'.format(size[0], size[1]) for size in args.size]

    if args.true == True:
        where += ['1']

    if args.type is not None:
        where += ['(type == "{}")'.format(args.type)]

    if args.uid is not None:
        where += ['(uid {} {})'.format(uid[0], uid[1]) for uid in args.uid]

    # if args.used is not None:

    if args.user is not None:
        where += ['uid == {}'.format(args.user)]

    # if args.wholename is not None:

    if args.writable == True:
        where += ['(mode & 128) == 128'] # 0200 have to use decimal or hexadecimal value because sqlite3 doesn't support octal

    # if args.xtype is not None:

    # if args.context is not None:

    return where

def build_group_by(args):
    '''Build the GROUP BY clause'''
    group_by = []

    return group_by

def build_order_by(args):
    '''Build the ORDER BY clause'''
    order_by = []

    if args.smallest:
        order_by += ['size ASC']

    if args.largest:
        order_by += ['size DESC']

    return order_by

def parse_escape(fmt, i):
    ret = fmt[i]

    if   ret == 'a':
        ret = '\a'
    elif ret == 'b':
        ret = '\b'
    elif ret == 'c':
        ret = ''
        i = len(fmt)
    elif ret == 'f':
        ret = '\f'
    elif ret == 'n':
        ret = '\n'
    elif ret == 'r':
        ret = '\r'
    elif ret == 't':
        ret = '\t'
    elif ret == 'v':
        ret = '\v'
    elif ret == '0':
        ret = '\0'
    elif ret == '\\':
        ret = '\\'
    elif ret.isdigit():
        chars = 1 + fmt[i + 1].isdigit() + fmt[i + 2].isdigit()
        ret = chr(int(fmt[i:chars + 1], 8))
        i += chars - 1
    else:
        pass

    return '"' + ret + '"', i + 1

def parse_directives(fmt, fmt_len, i, path):
    # unterminated directive prints ""; behavior is undefined in GNU find
    if i >= fmt_len:
        return '""', i

    # check for width formatting
    width = 0
    if fmt[i].isdigit() or (fmt[i] == '-') or (fmt[i] == '+') or (fmt[i] == '.'):
        j = i
        decimal_point = 0
        while fmt[j].isdigit() or (fmt[j] == '.'):
            if fmt[j] == '.':
                decimal_point += 1
            j += 1

        if decimal_point > 0:
            # let float deal with multiple decimal points
            width = float(fmt[i:j])
        else:
            width = int(fmt[i:j])

        i = j

    ret = fmt[i]

    if   ret == '%':
        ret = '%'
    elif ret == 'a':
        ret = 'printf("%{}d", atime)'
    # elif ret == 'A':
    elif ret == 'b':
        ret = 'printf("%{}f", (blocks * blksize) / 512)'
    elif ret == 'c':
        ret = 'printf("%{}d", ctime)'
    # elif ret == 'C':
    elif ret == 'd':
        ret = 'printf("%{}u", level())'
    elif ret == 'D':
        ret = 'printf("%{}s", "-")'
    elif ret == 'f':
        ret = 'printf("%{}s", name)'
    # elif ret == 'F':
    elif ret == 'g':
        ret = 'gidtogroup(gid, {})'
    elif ret == 'G':
        ret = 'printf("%{}d", gid)'
    elif ret == 'h':
        ret = 'printf("%{}s", ' + path + ')'
    elif ret == 'H':
        ret = 'printf("%{}s", starting_point())'
    elif ret == 'i':
        ret = 'printf("%{}d", inode)'
    elif ret == 'k':
        ret = 'printf("%{}f", (blocks * blksize) / 1024)'
    elif ret == 'l':
        ret = 'printf("%{}s", linkname)'
    elif ret == 'm':
        ret = 'printf("%{}o", mode & 0777)'
    elif ret == 'M':
        ret = 'printf("%{}s", modetotxt(mode))'
    elif ret == 'n':
        ret = 'printf("%{}d", nlink)'
    elif ret == 'p':
        ret = path + ' || (case type WHEN "d" THEN "" ELSE  "/" || name END)'
    elif ret == 'P':
        ret = 'printf("%{}s", substr(' + path + ' || "/" || name, length(starting_point()) + 2))'
    elif ret == 's':
        ret = 'printf("%{}u", size)'
    elif ret == 'S':
        ret = 'printf("%{}f", (blocksize * blocks) / size)'
    elif ret == 't':
        ret = 'printf("%{}d", mtime)'
    # elif ret == 'T':
    elif ret == 'u':
        ret = 'uidtouser(uid, {})'
    elif ret == 'U':
        ret = 'printf("%{}d", uid)'
    elif ret == 'y':
        ret = 'printf("%{}s", type)'
    # elif ret == 'Y':
    # elif ret == 'Z':
    else:
        ret = ''
        while (i < fmt_len) and (fmt[i] != '%') and (fmt[i] != '\\'):
            ret += fmt[i]
            i += 1
        ret = '"' + ret + '"'
        i -= 1

    return ret.format(width), i + 1

def build_output(args):
    '''Build the output columns SELECT clause'''

    output = ''

    path = 'path'
    if not need_aggregation(args):
        path += '()'

    if args.printf is not None:
        i = 0
        fmt = args.printf
        fmt_len = len(args.printf)

        output = '""'

        while i < fmt_len:
            output += ' || '
            if fmt[i] == '\\':
                column, i = parse_escape(fmt, i + 1)
                output += column
            elif fmt[i] == '%':
                column, i = parse_directives(fmt, fmt_len, i + 1, path)
                output += column
            else:
                output += '"' + fmt[i] + '"'
                i += 1;
    else:
        output = path + ' || (CASE type WHEN "d" THEN "" ELSE "/" || name END)'

    return [output]

def help(parser):
    # generate list of expressions
    expr = [expr for expr in sorted(parser.parse_args([]).__dict__.keys())]
    expr = [expr[i:i+5] for i in range(0, len(expr), 5)]

    print(('Usage: bee find [-P] [path...] [expression]\n' +
           '\n' +
           'default path is the current directory;\n' +
           'expression may consist of:\n' +
           '\n' +
           '\n'.join(['    ' + ' '.join(row) for row in expr]) + '\n' +
           '\n'))

    exit(0)

def build_expression_parser():
    # parse the arguments
    parser = argparse.ArgumentParser(description='BeeGFS Hive index version of find', add_help=False)

    # override help to not use -h
    parser.add_argument('-help', '--help',                                                         action='store_true',                                help='Print a summary of the command-line usage of find and exit.')

    # GNU find global options
    parser.add_argument('-maxdepth',           metavar='levels',     dest='maxdepth',              type=common.get_non_negative,                  help='Descend at most levels (a non-negative integer) levels of directories below the command line arguments. -maxdepth 0 means only apply the tests and actions to the command line arguments.')
    parser.add_argument('-mindepth',           metavar='levels',     dest='mindepth',              type=common.get_non_negative,                  help='Do not apply any tests or actions at levels less than levels (a non-negative integer). -mindepth 1 means process all files except the command line arguments.')
    parser.add_argument('--version', '-v',                                                         action='version',                                   version=os.path.basename(PATH) + ' 1.0.0')

    # GNU find test expressions
    parser.add_argument('-amin',               metavar='n',          dest='amin',                  type=numeric_arg, action='append',                  help='File was last accessed n minutes ago.')
    # parser.add_argument('-anewer',             metavar='file',       dest='anewer',                type=common.get_non_negative,                  help='File was last accessed more recently than file was modified.')
    parser.add_argument('-atime',              metavar='n',          dest='atime',                 type=numeric_arg, action='append',                  help='File was last accessed n*24 hours ago.')
    parser.add_argument('-cmin',               metavar='n',          dest='cmin',                  type=numeric_arg, action='append',                  help='File\'s status was last changed n minutes ago.')
    # parser.add_argument('-cnewer',             metavar='file',       dest='cnewer',                type=common.get_non_negative,                  help='File\'s status was last changed more recently than file was modified.')
    parser.add_argument('-ctime',              metavar='n',          dest='ctime',                 type=numeric_arg, action='append',                  help='File\'s status was last changed n*24 hours ago.')
    parser.add_argument('-empty',                                    dest='empty',                 action='store_true',                                help='File is empty and is either a regular file or a directory.')
    parser.add_argument('-executable',                               dest='executable',            action='store_true',                                help='Matches files which are executable and directories which are searchable (in a file name resolution sense).')
    parser.add_argument('-false',                                    dest='false',                 action='store_true',                                help='File is false and is either a regular file or a directory.')
    # parser.add_argument('-fstype',             metavar='n',          dest='fstype',                type=str,                                           help='File is on a filesystem of type type.')
    parser.add_argument('-gid',                metavar='n',          dest='gid',                   type=numeric_arg, action='append',                  help='File\'s numeric group ID is n.')
    parser.add_argument('-group',              metavar='gname',      dest='group',                 type=common.get_group,                         help='File belongs to group gname (numeric group ID allowed).')
    # parser.add_argument('-ilname',             metavar='pattern',    dest='ilname',                type=str,                                           help='Like -lname, but the match is case insensitive.')
    # parser.add_argument('-iname',              metavar='pattern',    dest='iname',                 type=str, action='append',                          help='Like -name, but the match is case insensitive.')
    parser.add_argument('-inum',               metavar='n',          dest='inum',                  type=numeric_arg, action='append',                  help='File has inode number n. It is normally easier to use the -samefile test instead.')
    # parser.add_argument('-ipath',              metavar='pattern',    dest='ipath',                 type=str,                                           help='Behaves in the same way as -iwholename. This option is deprecated, so please do not use it.')
    # parser.add_argument('-iregex',             metavar='pattern',    dest='iregex',                type=str,                                           help='Like -regex, but the match is case insensitive.')
    # parser.add_argument('-iwholename',         metavar='pattern',    dest='iwholename',            type=str,                                           help='Like -wholename, but the match is case insensitive.')
    parser.add_argument('-links',              metavar='n',          dest='links',                 type=numeric_arg, action='append',                  help='File has n links.')
    # parser.add_argument('-lname',              metavar='pattern',    dest='lname' ,                type=str,                                           help='File is a symbolic link whose contents match shell pattern pattern.')
    parser.add_argument('-mmin',               metavar='n',          dest='mmin',                  type=numeric_arg, action='append',                  help='File\'s data was last modified n minutes ago.')
    parser.add_argument('-mtime',              metavar='n',          dest='mtime',                 type=numeric_arg, action='append',                  help='File\'s data was last modified n*24 hours ago.')
    parser.add_argument('-name',               metavar='pattern',    dest='name',                  type=str, action='append',                          help='Base of file name (the path with the leading directories removed) matches shell pattern pattern.')
    parser.add_argument('-newer',              metavar='file',       dest='newer',                 type=os.lstat,                                      help='File was modified more recently than file.')
    # parser.add_argument('-newerXY',            metavar='reference',  dest='newerXY',               type=str,                                           help='Compares the timestamp of the current file with reference.')
    # parser.add_argument('-nouser',                                   dest='nouser',                type='store_true',                                  help='No user corresponds to file\'s numeric user ID.')
    parser.add_argument('-path',               metavar='pattern',    dest='path',                  type=str, action='append', default=[],              help='File name matches shell pattern pattern.')
    # parser.add_argument('-perm',               metavar='mode',       dest='perm',                  type=str,                                           help='File\'s  permission bits')
    parser.add_argument('-readable',                                 dest='readable',              action='store_true',                                help='Matches files which are readable.')
    # parser.add_argument('-regex',              metavar='pattern',    dest='regex',                 type=str,                                           help='File name matches regular expression pattern.')
    parser.add_argument('-samefile',           metavar='name',       dest='samefile',              type=os.lstat,                                      help='File refers to the same inode as name.')
    parser.add_argument('-size',               metavar='n',          dest='size',                  type=parse_size, action='append',                   help='')
    parser.add_argument('-true',                                     dest='true',                  action='store_true',                                help='Always true')
    parser.add_argument('-type',               metavar='c',          dest='type',                  type=common.get_char,                          help='File is of type c')
    parser.add_argument('-uid',                metavar='n',          dest='uid',                   type=numeric_arg,  action='append',                 help='File\'s numeric user ID is n.')
    # parser.add_argument('-used',               metavar='n',          dest='used',                  type=numeric_arg, action='append',                  help='File was last accessed n days after its status was last changed.')
    parser.add_argument('-user',               metavar='uname',      dest='user',                  type=common.get_user,                          help='File is owned by user uname (numeric user ID allowed).')
    # parser.add_argument('-wholename',          metavar='pattern',    dest='wholename',             type=str, action='append', default=[],              help='See -path. This alternative is less portable than -path.')
    parser.add_argument('-writable',                                 dest='writable',              action='store_true',                                help='Matches files which are writable.')
    # parser.add_argument('-xtype',              metavar='c',          dest='xtype',                 type=common.get_char,                          help='The same as -type unless the file is a symbolic link.')
    # parser.add_argument('-context',            metavar='pattern',    dest='context',               type=str,                                           help='(SELinux only) Security context of the file matches glob pattern.')

    # GNU find actions
    parser.add_argument('-fprint',             metavar='file',       dest='fprint',                type=str,                                           help='Output file prefix (Creates file <output>.tid)')
    parser.add_argument('-printf',             metavar='format',     dest='printf',                type=str,                                           help='print format on the standard output, similar to GNU find')

    # beegfs-hive-index specific arguments
    parser.add_argument('--num-results',       metavar='n',          dest='num_results',           type=common.get_non_negative,                  help='first n results')
    parser.add_argument('--smallest',                                dest='smallest',              action='store_true',                                help='top n smallest files')
    parser.add_argument('--largest',                                 dest='largest',               action='store_true',                                help='top n largest files')
    parser.add_argument('--in-memory-name',    metavar='name',       dest='inmemory_name',         type=str,                     default='out',        help='Name of in-memory database when aggregation is performed')

    return parser

# argv[0] should be the command name
def exec_find(argv, config_path):

    opt = ''

    # storing name of test config file
    test_config_path ="config.test"

    # find and parse the configuration file first
    if(config_path == bee_config.DEFAULT_PATH or config_path == test_config_path):
        config = bee_config.Server(config_path)
    else:
        config = bee_config.envServer(bee_config.DEFAULT_PATH)

    expression_parser = build_expression_parser()

    i = 2
    at_expressions = False # whether or not the command line arguments started with expressions

    # parse 'real' options
    # none are actually handled
    real = {
        # 'H': False,
        # 'L': False,
        'P': True,
        # 'D': False, # don't even handle here
        # 'O': False, # don't even handle here
    }

    while (i < len(argv)) and (argv[i][0] == '-'):
        if '-help' in argv[i]:
            help(expression_parser)

        opt = argv[i]

        # must be separate arguments
        # if any flags are not H, L, or P, immediately treat the rest of the flags as expression flags
        if argv[i][1:] not in ['H', 'L', 'P']:
            at_expressions = True
            break

        if argv[i][1:] in ['H', 'L']:
            sys.stderr.write('Found flag {}. Ignoring.\n'.format(argv[i][1:]))
        else:
            real[argv[i][1:]] = True

        i += 1

    if at_expressions and (argv[i][0] != '-'):
        raise Exception('bee find: paths must preceed expression: {}'.format(argv[i - 1]))

    # parse paths
    paths = []
    path_list = []
    indexpaths = []

    while (i < len(argv)) and (argv[i][0] != '-'):
        paths += [argv[i]]
        i += 1

    if len(paths) == 0:
        paths = ['']

    if(opt == '-P'):
        index_path = ''.join(''.join(paths))

       
    # prepend the provided paths with the beegfs-hive-index root path
    config_paths = config.indexpaths().split(':')
    fs_path = config_paths[0]
    if len(config_paths) == 2:
        index_path = config_paths[1]
    elif opt != '-P':
        index_path = config_paths[0]
    
    for path in paths:
        if len(path) == 0 or path[0] != "/":
            current_dir = os.getcwd()
            path1 = "%s/%s" % (current_dir, path)
        else:
            path1 = path

        # check if path is part of index or FSpath
        if not path1.startswith(index_path) and not path1.startswith(fs_path):
            print("Error: Invalid path. Please run from index or FSpath directory")
            exit(1)

        # If the path starts with fspath, replace it with index
        if path1.startswith(fs_path):
            path1 = path1.replace(fs_path, index_path)

        path_list += [path1]
        
    # parse expressions without the 'real' and path arguments
    args, unknown = expression_parser.parse_known_args(argv[i:])

    if args.help:
        help(expression_parser)

    if len(unknown):
        raise Exception('bee find: unknown predicate `{}\''.format(unknown[0]))

    validate_args(args)

    # generate the query clauses
    where    = build_where(args)
    group_by = build_group_by(args)
    order_by = build_order_by(args)
    output   = build_output(args)

    # create the query command
    exec_path = "%s/bin/index/bee-search" % config.executable()
    query_cmd = [exec_path,
                 '-n', str(config.threads()),
                 '-B', str(config.outputbuffer())
                ]

    S = None
    E = None

    # aggregation is required
    if need_aggregation(args):
        columns = ['name', 'type', 'inode', 'mode', 'nlink', 'uid',
                   'gid', 'size', 'blksize', 'blocks', 'atime',
                   'mtime', 'ctime', 'linkname', 'xattrs']

        I = 'CREATE TABLE {0} ({1})'.format(args.inmemory_name,
                                            ', '.join(['id INTEGER PRIMARY KEY',
                                                       'path TEXT',
                                                       'name TEXT',     'type TEXT',
                                                       'inode TEXT',    'mode INT64',
                                                       'nlink INT64',   'uid INT64',
                                                       'gid INT64',     'size INT64',
                                                       'blksize INT64', 'blocks INT64',
                                                       'atime INT64',   'mtime INT64',
                                                       'ctime INT64',   'linkname TEXT',
                                                       'xattrs TEXT']))

        S = 'INSERT INTO {0} {1}'.format(args.inmemory_name,
                                         common.build_query(['NULL', 'path()'] + columns,
                                                                 ['summary'],
                                                                 where,
                                                                 group_by,
                                                                 order_by,
                                                                 args.num_results))

        E = 'INSERT INTO {0} {1}'.format(args.inmemory_name,
                                         common.build_query(['NULL', 'path()'] + columns,
                                                                 ['pentries'],
                                                                 where,
                                                                 group_by,
                                                                 order_by,
                                                                 args.num_results))

        J = "INSERT INTO aggregate.{0} {1}".format(args.inmemory_name,
                                                   common.build_query(['NULL', 'path'] + columns,
                                                                           [args.inmemory_name],
                                                                           where,
                                                                           group_by,
                                                                           order_by,
                                                                           args.num_results))
        G = common.build_query(output,
                                    [args.inmemory_name],
                                    where,
                                    group_by,
                                    order_by,
                                    args.num_results)

        query_cmd += ['-e', '0',
                      '-I', I,
                      '-J', J,
                      '-G', G]
    else:
        S = common.build_query(output,
                                    ['summary'],
                                    where,
                                    group_by,
                                    order_by,
                                    args.num_results)

        E = common.build_query(output,
                                    ['pentries'],
                                    where,
                                    group_by,
                                    order_by,
                                    args.num_results)

    query_cmd += ['-S', S,
                  '-E', E,
                  '-a',
                  '-d', ' ']

    if args.maxdepth is not None:
        query_cmd += ['-z', str(args.maxdepth)]

    if args.mindepth is not None:
        query_cmd += ['-y', str(args.mindepth)]

    if args.fprint:
        query_cmd += ['-o', args.fprint]

    query = subprocess.Popen(query_cmd + path_list + [fs_path]) # positional arguments must appear after flags

    query.communicate()                         # block until query finishes

    return query.returncode
